@@ -36,7 +36,7 @@
 
 #define SYMBOL_QUOTE		((xmlChar)'\'')
 
-#define DEFAULT_TOKEN		'0'
+#define DEFAULT_TOKEN		(xmlChar)'0'
 #define DEFAULT_SEPARATOR	"."
 
 #define MAX_TOKENS		1024
@@ -45,7 +45,7 @@ typedef struct _xsltFormatToken xsltFormatToken;
 typedef xsltFormatToken *xsltFormatTokenPtr;
 struct _xsltFormatToken {
     xmlChar	*separator;
-    int		 token;
+    xmlChar	 token;
     int		 width;
 };
 
@@ -65,11 +65,39 @@ static xsltFormatToken default_token;
 /*
  * **** Start temp insert ****
  *
- * The following routine xsltUTF8Charcmp will be replaced with calls to
- * the corresponding libxml routine at a later date (when other
- * inter-library dependencies require it).
+ * The following two routines (xsltUTF8Size and xsltUTF8Charcmp)
+ * will be replaced with calls to the corresponding libxml routines
+ * at a later date (when other inter-library dependencies require it)
  */
 
+/**
+ * xsltUTF8Size:
+ * @utf: pointer to the UTF8 character
+ *
+ * returns the numbers of bytes in the character, -1 on format error
+ */
+static int
+xsltUTF8Size(xmlChar *utf) {
+    xmlChar mask;
+    int len;
+
+    if (utf == NULL)
+        return -1;
+    if (*utf < 0x80)
+        return 1;
+    /* check valid UTF8 character */
+    if (!(*utf & 0x40))
+        return -1;
+    /* determine number of bytes in char */
+    len = 2;
+    for (mask=0x20; mask != 0; mask>>=1) {
+        if (!(*utf & mask))
+            return len;
+        len++;
+    }
+    return -1;
+}
+
 /**
  * xsltUTF8Charcmp
  * @utf1: pointer to first UTF8 char
@@ -80,16 +108,13 @@ static xsltFormatToken default_token;
  */
 static int
 xsltUTF8Charcmp(xmlChar *utf1, xmlChar *utf2) {
-    int len = xmlUTF8Strsize(utf1, 1);
 
-    if (len < 1)
-        return -1;
     if (utf1 == NULL ) {
         if (utf2 == NULL)
             return 0;
         return -1;
     }
-    return xmlStrncmp(utf1, utf2, len);
+    return xmlStrncmp(utf1, utf2, xsltUTF8Size(utf1));
 }
 
 /***** Stop temp insert *****/
@@ -107,22 +132,20 @@ xsltUTF8Charcmp(xmlChar *utf1, xmlChar *utf2) {
      (xsltUTF8Charcmp((letter), (self)->patternSeparator) == 0))
 
 #define IS_DIGIT_ZERO(x) xsltIsDigitZero(x)
-#define IS_DIGIT_ONE(x) xsltIsDigitZero((x)-1)
+#define IS_DIGIT_ONE(x) xsltIsDigitZero((xmlChar)(x)-1)
 
 static int
 xsltIsDigitZero(unsigned int ch)
 {
     /*
      * Reference: ftp://ftp.unicode.org/Public/UNIDATA/UnicodeData.txt
-     *
-     * There a many more digit ranges in newer Unicode versions. These
-     * are only the zeros that match Digit in XML 1.0 (IS_DIGIT macro).
      */
     switch (ch) {
     case 0x0030: case 0x0660: case 0x06F0: case 0x0966:
     case 0x09E6: case 0x0A66: case 0x0AE6: case 0x0B66:
     case 0x0C66: case 0x0CE6: case 0x0D66: case 0x0E50:
-    case 0x0ED0: case 0x0F20:
+    case 0x0E60: case 0x0F20: case 0x1040: case 0x17E0:
+    case 0x1810: case 0xFF10:
 	return TRUE;
     default:
 	return FALSE;
@@ -170,7 +193,7 @@ xsltNumberFormatDecimal(xmlBufferPtr buffer,
 	    pointer -= groupingCharacterLen;
 	    xmlCopyCharMultiByte(pointer, groupingCharacter);
 	}
-
+	
 	val = digit_zero + (int)fmod(number, 10.0);
 	if (val < 0x80) {			/* shortcut if ASCII */
 	    if (pointer <= temp_string) {	/* Check enough room */
@@ -180,7 +203,7 @@ xsltNumberFormatDecimal(xmlBufferPtr buffer,
 	    *(--pointer) = val;
 	}
 	else {
-	/*
+	/* 
 	 * Here we have a multibyte character.  It's a little messy,
 	 * because until we generate the char we don't know how long
 	 * it is.  So, we generate it into the buffer temp_char, then
@@ -199,13 +222,12 @@ xsltNumberFormatDecimal(xmlBufferPtr buffer,
     }
     if (i < 0)
         xsltGenericError(xsltGenericErrorContext,
-		"xsltNumberFormatDecimal: Internal buffer size exceeded\n");
+		"xsltNumberFormatDecimal: Internal buffer size exceeded");
     xmlBufferCat(buffer, pointer);
 }
 
 static void
-xsltNumberFormatAlpha(xsltNumberDataPtr data,
-		      xmlBufferPtr buffer,
+xsltNumberFormatAlpha(xmlBufferPtr buffer,
 		      double number,
 		      int is_upper)
 {
@@ -215,59 +237,26 @@ xsltNumberFormatAlpha(xsltNumberDataPtr data,
     char *alpha_list;
     double alpha_size = (double)(sizeof(alpha_upper_list) - 1);
 
-    /*
-     * XSLT 1.0 isn't clear on how to handle zero, but XSLT 2.0 says:
-     *
-     *     For all format tokens other than the first kind above (one that
-     *     consists of decimal digits), there may be implementation-defined
-     *     lower and upper bounds on the range of numbers that can be
-     *     formatted using this format token; indeed, for some numbering
-     *     sequences there may be intrinsic limits. [...] Numbers that fall
-     *     outside this range must be formatted using the format token 1.
-     *
-     * The "a" token has an intrinsic lower limit of 1.
-     */
-    if (number < 1.0) {
-        xsltNumberFormatDecimal(buffer, number, '0', 1,
-                                data->digitsPerGroup,
-                                data->groupingCharacter,
-                                data->groupingCharacterLen);
-        return;
-    }
-
     /* Build buffer from back */
     pointer = &temp_string[sizeof(temp_string)];
     *(--pointer) = 0;
     alpha_list = (is_upper) ? alpha_upper_list : alpha_lower_list;
-
+    
     for (i = 1; i < (int)sizeof(temp_string); i++) {
 	number--;
 	*(--pointer) = alpha_list[((int)fmod(number, alpha_size))];
 	number /= alpha_size;
-	if (number < 1.0)
+	if (fabs(number) < 1.0)
 	    break; /* for */
     }
     xmlBufferCCat(buffer, pointer);
 }
 
 static void
-xsltNumberFormatRoman(xsltNumberDataPtr data,
-		      xmlBufferPtr buffer,
+xsltNumberFormatRoman(xmlBufferPtr buffer,
 		      double number,
 		      int is_upper)
 {
-    /*
-     * See discussion in xsltNumberFormatAlpha. Also use a reasonable upper
-     * bound to avoid denial of service.
-     */
-    if (number < 1.0 || number > 5000.0) {
-        xsltNumberFormatDecimal(buffer, number, '0', 1,
-                                data->digitsPerGroup,
-                                data->groupingCharacter,
-                                data->groupingCharacterLen);
-        return;
-    }
-
     /*
      * Based on an example by Jim Walsh
      */
@@ -329,7 +318,7 @@ static void
 xsltNumberFormatTokenize(const xmlChar *format,
 			 xsltFormatPtr tokens)
 {
-    int ix = 0;
+    int index = 0;
     int j;
     int val;
     int len;
@@ -347,19 +336,19 @@ xsltNumberFormatTokenize(const xmlChar *format,
      * Insert initial non-alphanumeric token.
      * There is always such a token in the list, even if NULL
      */
-    while (! (IS_LETTER(val=xmlStringCurrentChar(NULL, format+ix, &len)) ||
-	      IS_DIGIT(val)) ) {
-	if (format[ix] == 0)		/* if end of format string */
+    while (! (IS_LETTER(val=xmlStringCurrentChar(NULL, format+index, &len)) ||
+    	      IS_DIGIT(val)) ) {
+	if (format[index] == 0)		/* if end of format string */
 	    break; /* while */
-	ix += len;
+	index += len;
     }
-    if (ix > 0)
-	tokens->start = xmlStrndup(format, ix);
+    if (index > 0)
+	tokens->start = xmlStrndup(format, index);
 
 
     for (tokens->nTokens = 0; tokens->nTokens < MAX_TOKENS;
 	 tokens->nTokens++) {
-	if (format[ix] == 0)
+	if (format[index] == 0)
 	    break; /* for */
 
 	/*
@@ -371,30 +360,27 @@ xsltNumberFormatTokenize(const xmlChar *format,
 	    tokens->end = NULL;
 	}
 
-	val = xmlStringCurrentChar(NULL, format+ix, &len);
+	val = xmlStringCurrentChar(NULL, format+index, &len);
 	if (IS_DIGIT_ONE(val) ||
 		 IS_DIGIT_ZERO(val)) {
 	    tokens->tokens[tokens->nTokens].width = 1;
 	    while (IS_DIGIT_ZERO(val)) {
 		tokens->tokens[tokens->nTokens].width++;
-		ix += len;
-		val = xmlStringCurrentChar(NULL, format+ix, &len);
+		index += len;
+		val = xmlStringCurrentChar(NULL, format+index, &len);
 	    }
 	    if (IS_DIGIT_ONE(val)) {
 		tokens->tokens[tokens->nTokens].token = val - 1;
-		ix += len;
-		val = xmlStringCurrentChar(NULL, format+ix, &len);
-	    } else {
-                tokens->tokens[tokens->nTokens].token = '0';
-                tokens->tokens[tokens->nTokens].width = 1;
-            }
-	} else if ( (val == 'A') ||
-		    (val == 'a') ||
-		    (val == 'I') ||
-		    (val == 'i') ) {
+		index += len;
+		val = xmlStringCurrentChar(NULL, format+index, &len);
+	    }
+	} else if ( (val == (xmlChar)'A') ||
+		    (val == (xmlChar)'a') ||
+		    (val == (xmlChar)'I') ||
+		    (val == (xmlChar)'i') ) {
 	    tokens->tokens[tokens->nTokens].token = val;
-	    ix += len;
-	    val = xmlStringCurrentChar(NULL, format+ix, &len);
+	    index += len;
+	    val = xmlStringCurrentChar(NULL, format+index, &len);
 	} else {
 	    /* XSLT section 7.7
 	     * "Any other format token indicates a numbering sequence
@@ -402,7 +388,7 @@ xsltNumberFormatTokenize(const xmlChar *format,
 	     *  not support a numbering sequence that starts with that
 	     *  token, it must use a format token of 1."
 	     */
-	    tokens->tokens[tokens->nTokens].token = '0';
+	    tokens->tokens[tokens->nTokens].token = (xmlChar)'0';
 	    tokens->tokens[tokens->nTokens].width = 1;
 	}
 	/*
@@ -415,22 +401,22 @@ xsltNumberFormatTokenize(const xmlChar *format,
 	 * one wonders why XSLT doesn't refer to these instead).
 	 */
 	while (IS_LETTER(val) || IS_DIGIT(val)) {
-	    ix += len;
-	    val = xmlStringCurrentChar(NULL, format+ix, &len);
+	    index += len;
+	    val = xmlStringCurrentChar(NULL, format+index, &len);
 	}
 
 	/*
 	 * Insert temporary non-alphanumeric final tooken.
 	 */
-	j = ix;
+	j = index;
 	while (! (IS_LETTER(val) || IS_DIGIT(val))) {
 	    if (val == 0)
 		break; /* while */
-	    ix += len;
-	    val = xmlStringCurrentChar(NULL, format+ix, &len);
+	    index += len;
+	    val = xmlStringCurrentChar(NULL, format+index, &len);
 	}
-	if (ix > j)
-	    tokens->end = xmlStrndup(&format[j], ix - j);
+	if (index > j)
+	    tokens->end = xmlStrndup(&format[j], index - j);
     }
 }
 
@@ -454,23 +440,6 @@ xsltNumberFormatInsertNumbers(xsltNumberDataPtr data,
     for (i = 0; i < numbers_max; i++) {
 	/* Insert number */
 	number = numbers[(numbers_max - 1) - i];
-        /* Round to nearest like XSLT 2.0 */
-        number = floor(number + 0.5);
-        /*
-         * XSLT 1.0 isn't clear on how to handle negative numbers, but XSLT
-         * 2.0 says:
-         *
-         *     It is a non-recoverable dynamic error if any undiscarded item
-         *     in the atomized sequence supplied as the value of the value
-         *     attribute of xsl:number cannot be converted to an integer, or
-         *     if the resulting integer is less than 0 (zero).
-         */
-        if (number < 0.0) {
-            xsltTransformError(NULL, NULL, NULL,
-                    "xsl-number : negative value\n");
-            /* Recover by treating negative values as zero. */
-            number = 0.0;
-        }
 	if (i < tokens->nTokens) {
 	  /*
 	   * The "n"th format token will be used to format the "n"th
@@ -514,16 +483,28 @@ xsltNumberFormatInsertNumbers(xsltNumberDataPtr data,
 
 		switch (token->token) {
 		case 'A':
-		    xsltNumberFormatAlpha(data, buffer, number, TRUE);
+		    xsltNumberFormatAlpha(buffer,
+					  number,
+					  TRUE);
+
 		    break;
 		case 'a':
-		    xsltNumberFormatAlpha(data, buffer, number, FALSE);
+		    xsltNumberFormatAlpha(buffer,
+					  number,
+					  FALSE);
+
 		    break;
 		case 'I':
-		    xsltNumberFormatRoman(data, buffer, number, TRUE);
+		    xsltNumberFormatRoman(buffer,
+					  number,
+					  TRUE);
+
 		    break;
 		case 'i':
-		    xsltNumberFormatRoman(data, buffer, number, FALSE);
+		    xsltNumberFormatRoman(buffer,
+					  number,
+					  FALSE);
+
 		    break;
 		default:
 		    if (IS_DIGIT_ZERO(token->token)) {
@@ -550,59 +531,62 @@ xsltNumberFormatInsertNumbers(xsltNumberDataPtr data,
 
 }
 
-static int
-xsltTestCompMatchCount(xsltTransformContextPtr context,
-                       xmlNodePtr node,
-                       xsltCompMatchPtr countPat,
-                       xmlNodePtr cur)
-{
-    if (countPat != NULL) {
-        return xsltTestCompMatchList(context, node, countPat);
-    }
-    else {
-        /*
-         * 7.7 Numbering
-         *
-         * If count attribute is not specified, then it defaults to the
-         * pattern that matches any node with the same node type as the
-         * current node and, if the current node has an expanded-name, with
-         * the same expanded-name as the current node.
-         */
-        if (node->type != cur->type)
-            return 0;
-        if (node->type == XML_NAMESPACE_DECL)
-            /*
-             * Namespace nodes have no preceding siblings and no parents
-             * that are namespace nodes. This means that node == cur.
-             */
-            return 1;
-        /* TODO: Skip node types without expanded names like text nodes. */
-        if (!xmlStrEqual(node->name, cur->name))
-            return 0;
-        if (node->ns == cur->ns)
-            return 1;
-        if ((node->ns == NULL) || (cur->ns == NULL))
-            return 0;
-        return (xmlStrEqual(node->ns->href, cur->ns->href));
-    }
-}
-
 static int
 xsltNumberFormatGetAnyLevel(xsltTransformContextPtr context,
 			    xmlNodePtr node,
-			    xsltCompMatchPtr countPat,
-			    xsltCompMatchPtr fromPat,
-			    double *array)
+			    const xmlChar *count,
+			    const xmlChar *from,
+			    double *array,
+			    xmlDocPtr doc,
+			    xmlNodePtr elem)
 {
     int amount = 0;
     int cnt = 0;
-    xmlNodePtr cur = node;
+    xmlNodePtr cur;
+    xsltCompMatchPtr countPat = NULL;
+    xsltCompMatchPtr fromPat = NULL;
+
+    if (count != NULL)
+	countPat = xsltCompilePattern(count, doc, elem, NULL, context);
+    if (from != NULL)
+	fromPat = xsltCompilePattern(from, doc, elem, NULL, context);
+	
+    /* select the starting node */
+    switch (node->type) {
+	case XML_ELEMENT_NODE:
+	    cur = node;
+	    break;
+	case XML_ATTRIBUTE_NODE:
+	    cur = ((xmlAttrPtr) node)->parent;
+	    break;
+	case XML_TEXT_NODE:
+	case XML_PI_NODE:
+	case XML_COMMENT_NODE:
+	    cur = node->parent;
+	    break;
+	default:
+	    cur = NULL;
+	    break;
+    }
 
     while (cur != NULL) {
 	/* process current node */
-	if (xsltTestCompMatchCount(context, cur, countPat, node))
-	    cnt++;
-	if ((fromPat != NULL) &&
+	if (count == NULL) {
+	    if ((node->type == cur->type) &&
+		/* FIXME: must use expanded-name instead of local name */
+		xmlStrEqual(node->name, cur->name)) {
+		    if ((node->ns == cur->ns) ||
+		        ((node->ns != NULL) &&
+			 (cur->ns != NULL) &&
+		         (xmlStrEqual(node->ns->href,
+		             cur->ns->href) )))
+		        cnt++;
+	    }
+	} else {
+	    if (xsltTestCompMatchList(context, cur, countPat))
+		cnt++;
+	}
+	if ((from != NULL) &&
 	    xsltTestCompMatchList(context, cur, fromPat)) {
 	    break; /* while */
 	}
@@ -615,87 +599,99 @@ xsltNumberFormatGetAnyLevel(xsltTransformContextPtr context,
             (cur->type == XML_HTML_DOCUMENT_NODE))
 	    break; /* while */
 
-        if (cur->type == XML_NAMESPACE_DECL) {
-            /*
-            * The XPath module stores the parent of a namespace node in
-            * the ns->next field.
-            */
-            cur = (xmlNodePtr) ((xmlNsPtr) cur)->next;
-        } else if (cur->type == XML_ATTRIBUTE_NODE) {
-            cur = cur->parent;
-        } else {
-            while ((cur->prev != NULL) && ((cur->prev->type == XML_DTD_NODE) ||
-                   (cur->prev->type == XML_XINCLUDE_START) ||
-                   (cur->prev->type == XML_XINCLUDE_END)))
-                cur = cur->prev;
-            if (cur->prev != NULL) {
-                for (cur = cur->prev; cur->last != NULL; cur = cur->last);
-            } else {
-                cur = cur->parent;
-            }
-        }
+	while ((cur->prev != NULL) && ((cur->prev->type == XML_DTD_NODE) ||
+	       (cur->prev->type == XML_XINCLUDE_START) ||
+	       (cur->prev->type == XML_XINCLUDE_END)))
+	    cur = cur->prev;
+	if (cur->prev != NULL) {
+	    for (cur = cur->prev; cur->last != NULL; cur = cur->last);
+	} else {
+	    cur = cur->parent;
+	}
+
     }
 
     array[amount++] = (double) cnt;
 
+    if (countPat != NULL)
+	xsltFreeCompMatchList(countPat);
+    if (fromPat != NULL)
+	xsltFreeCompMatchList(fromPat);
     return(amount);
 }
 
 static int
 xsltNumberFormatGetMultipleLevel(xsltTransformContextPtr context,
 				 xmlNodePtr node,
-				 xsltCompMatchPtr countPat,
-				 xsltCompMatchPtr fromPat,
+				 const xmlChar *count,
+				 const xmlChar *from,
 				 double *array,
-				 int max)
+				 int max,
+				 xmlDocPtr doc,
+				 xmlNodePtr elem)
 {
     int amount = 0;
     int cnt;
-    xmlNodePtr oldCtxtNode;
     xmlNodePtr ancestor;
     xmlNodePtr preceding;
     xmlXPathParserContextPtr parser;
+    xsltCompMatchPtr countPat;
+    xsltCompMatchPtr fromPat;
 
-    oldCtxtNode = context->xpathCtxt->node;
+    if (count != NULL)
+	countPat = xsltCompilePattern(count, doc, elem, NULL, context);
+    else
+	countPat = NULL;
+    if (from != NULL)
+	fromPat = xsltCompilePattern(from, doc, elem, NULL, context);
+    else
+	fromPat = NULL;
+    context->xpathCtxt->node = node;
     parser = xmlXPathNewParserContext(NULL, context->xpathCtxt);
     if (parser) {
 	/* ancestor-or-self::*[count] */
-	ancestor = node;
-	while ((ancestor != NULL) && (ancestor->type != XML_DOCUMENT_NODE)) {
-	    if ((fromPat != NULL) &&
+	for (ancestor = node;
+	     (ancestor != NULL) && (ancestor->type != XML_DOCUMENT_NODE);
+	     ancestor = xmlXPathNextAncestor(parser, ancestor)) {
+	    
+	    if ((from != NULL) &&
 		xsltTestCompMatchList(context, ancestor, fromPat))
 		break; /* for */
-
-            /*
-             * The xmlXPathNext* iterators require that the context node is
-             * set to the start node. Calls to xsltTestCompMatch* may also
-             * leave the context node in an undefined state, so make sure
-             * that the context node is reset before each iterator invocation.
-             */
-
-	    if (xsltTestCompMatchCount(context, ancestor, countPat, node)) {
+	    
+	    if ((count == NULL && node->type == ancestor->type && 
+		xmlStrEqual(node->name, ancestor->name)) ||
+		xsltTestCompMatchList(context, ancestor, countPat)) {
 		/* count(preceding-sibling::*) */
-		cnt = 1;
-                context->xpathCtxt->node = ancestor;
-                preceding = xmlXPathNextPrecedingSibling(parser, ancestor);
-                while (preceding != NULL) {
-	            if (xsltTestCompMatchCount(context, preceding, countPat,
-                                               node))
-			cnt++;
-                    context->xpathCtxt->node = ancestor;
-                    preceding =
-                        xmlXPathNextPrecedingSibling(parser, preceding);
+		cnt = 0;
+		for (preceding = ancestor;
+		     preceding != NULL;
+		     preceding = 
+		        xmlXPathNextPrecedingSibling(parser, preceding)) {
+		    if (count == NULL) {
+			if ((preceding->type == ancestor->type) &&
+			    xmlStrEqual(preceding->name, ancestor->name)){
+			    if ((preceding->ns == ancestor->ns) ||
+			        ((preceding->ns != NULL) &&
+				 (ancestor->ns != NULL) &&
+			         (xmlStrEqual(preceding->ns->href,
+			             ancestor->ns->href) )))
+			        cnt++;
+			}
+		    } else {
+			if (xsltTestCompMatchList(context, preceding,
+				                  countPat))
+			    cnt++;
+		    }
 		}
 		array[amount++] = (double)cnt;
 		if (amount >= max)
 		    break; /* for */
 	    }
-            context->xpathCtxt->node = node;
-            ancestor = xmlXPathNextAncestor(parser, ancestor);
 	}
 	xmlXPathFreeParserContext(parser);
     }
-    context->xpathCtxt->node = oldCtxtNode;
+    xsltFreeCompMatchList(countPat);
+    xsltFreeCompMatchList(fromPat);
     return amount;
 }
 
@@ -708,7 +704,7 @@ xsltNumberFormatGetValue(xmlXPathContextPtr context,
     int amount = 0;
     xmlBufferPtr pattern;
     xmlXPathObjectPtr obj;
-
+    
     pattern = xmlBufferCreate();
     if (pattern != NULL) {
 	xmlBufferCCat(pattern, "number(");
@@ -730,7 +726,7 @@ xsltNumberFormatGetValue(xmlXPathContextPtr context,
 /**
  * xsltNumberFormat:
  * @ctxt: the XSLT transformation context
- * @data: the formatting information
+ * @data: the formatting informations
  * @node: the data to format
  *
  * Convert one number.
@@ -744,29 +740,24 @@ xsltNumberFormat(xsltTransformContextPtr ctxt,
     int amount, i;
     double number;
     xsltFormat tokens;
+    int tempformat = 0;
 
-    if (data->format != NULL) {
-        xsltNumberFormatTokenize(data->format, &tokens);
-    }
-    else {
-        xmlChar *format;
-
-	/* The format needs to be recomputed each time */
-        if (data->has_format == 0)
-            return;
-	format = xsltEvalAttrValueTemplate(ctxt, data->node,
+    if ((data->format == NULL) && (data->has_format != 0)) {
+	data->format = xsltEvalAttrValueTemplate(ctxt, data->node,
 					     (const xmlChar *) "format",
 					     XSLT_NAMESPACE);
-        if (format == NULL)
-            return;
-        xsltNumberFormatTokenize(format, &tokens);
-	xmlFree(format);
+	tempformat = 1;
+    }
+    if (data->format == NULL) {
+	return;
     }
 
     output = xmlBufferCreate();
     if (output == NULL)
 	goto XSLT_NUMBER_FORMAT_END;
 
+    xsltNumberFormatTokenize(data->format, &tokens);
+
     /*
      * Evaluate the XPath expression to find the value(s)
      */
@@ -782,16 +773,18 @@ xsltNumberFormat(xsltTransformContextPtr ctxt,
 					  &tokens,
 					  output);
 	}
-
+	
     } else if (data->level) {
-
+	
 	if (xmlStrEqual(data->level, (const xmlChar *) "single")) {
 	    amount = xsltNumberFormatGetMultipleLevel(ctxt,
 						      node,
-						      data->countPat,
-						      data->fromPat,
+						      data->count,
+						      data->from,
 						      &number,
-						      1);
+						      1,
+						      data->doc,
+						      data->node);
 	    if (amount == 1) {
 		xsltNumberFormatInsertNumbers(data,
 					      &number,
@@ -804,10 +797,12 @@ xsltNumberFormat(xsltTransformContextPtr ctxt,
 	    int max = sizeof(numarray)/sizeof(numarray[0]);
 	    amount = xsltNumberFormatGetMultipleLevel(ctxt,
 						      node,
-						      data->countPat,
-						      data->fromPat,
+						      data->count,
+						      data->from,
 						      numarray,
-						      max);
+						      max,
+						      data->doc,
+						      data->node);
 	    if (amount > 0) {
 		xsltNumberFormatInsertNumbers(data,
 					      numarray,
@@ -818,9 +813,11 @@ xsltNumberFormat(xsltTransformContextPtr ctxt,
 	} else if (xmlStrEqual(data->level, (const xmlChar *) "any")) {
 	    amount = xsltNumberFormatGetAnyLevel(ctxt,
 						 node,
-						 data->countPat,
-						 data->fromPat,
-						 &number);
+						 data->count,
+						 data->from,
+						 &number, 
+						 data->doc,
+						 data->node);
 	    if (amount > 0) {
 		xsltNumberFormatInsertNumbers(data,
 					      &number,
@@ -829,23 +826,10 @@ xsltNumberFormat(xsltTransformContextPtr ctxt,
 					      output);
 	    }
 	}
-
-        /*
-         * Unlike `match` patterns, `count` and `from` patterns can contain
-         * variable references, so we have to clear the pattern match
-         * cache if the "direct" matching algorithm was used.
-         */
-        if (data->countPat != NULL)
-            xsltCompMatchClearCache(ctxt, data->countPat);
-        if (data->fromPat != NULL)
-            xsltCompMatchClearCache(ctxt, data->fromPat);
     }
     /* Insert number as text node */
     xsltCopyTextString(ctxt, ctxt->insert, xmlBufferContent(output), 0);
 
-    xmlBufferFree(output);
-
-XSLT_NUMBER_FORMAT_END:
     if (tokens.start != NULL)
 	xmlFree(tokens.start);
     if (tokens.end != NULL)
@@ -854,19 +838,26 @@ XSLT_NUMBER_FORMAT_END:
 	if (tokens.tokens[i].separator != NULL)
 	    xmlFree(tokens.tokens[i].separator);
     }
+    
+XSLT_NUMBER_FORMAT_END:
+    if (tempformat == 1) {
+	/* The format need to be recomputed each time */
+	data->format = NULL;
+    }
+    if (output != NULL)
+	xmlBufferFree(output);
 }
 
 static int
 xsltFormatNumberPreSuffix(xsltDecimalFormatPtr self, xmlChar **format, xsltFormatNumberInfoPtr info)
 {
-    /* will hold total length of prefix/suffix without quote characters */
-    int	count=0;
+    int	count=0;	/* will hold total length of prefix/suffix */
     int len;
 
     while (1) {
-	/*
-	 * prefix / suffix ends at end of string or at
-	 * first 'special' character
+	/* 
+	 * prefix / suffix ends at end of string or at 
+	 * first 'special' character 
 	 */
 	if (**format == 0)
 	    return count;
@@ -874,17 +865,22 @@ xsltFormatNumberPreSuffix(xsltDecimalFormatPtr self, xmlChar **format, xsltForma
 	if (**format == SYMBOL_QUOTE) {
 	    if (*++(*format) == 0)
 		return -1;
+	    if ((len=xsltUTF8Size(*format))<=0)
+	        return -1;
+	    *format += len;
+	    if (**format != SYMBOL_QUOTE)
+		return -1;
 	}
 	else if (IS_SPECIAL(self, *format))
 	    return count;
 	/*
 	 * else treat percent/per-mille as special cases,
-	 * depending on whether +ve or -ve
+	 * depending on whether +ve or -ve 
 	 */
-	else {
+	else if (!info->is_negative_pattern) {
 	    /*
-	     * for +ve prefix/suffix, allow only a
-	     * single occurence of either
+	     * for +ve prefix/suffix, allow only a 
+	     * single occurence of either 
 	     */
 	    if (xsltUTF8Charcmp(*format, self->percent) == 0) {
 		if (info->is_multiplier_set)
@@ -897,21 +893,39 @@ xsltFormatNumberPreSuffix(xsltDecimalFormatPtr self, xmlChar **format, xsltForma
 		info->multiplier = 1000;
 		info->is_multiplier_set = TRUE;
 	    }
+	} else {
+	    /*
+	     * for -ve prefix/suffix, allow only single occurence 
+	     * & insist it's previously defined 
+	     */
+	    if (xsltUTF8Charcmp(*format, self->percent) == 0) {
+		if (info->is_multiplier_set)
+		    return -1;
+		if (info->multiplier != 100)
+		    return -1;
+		info->is_multiplier_set = TRUE;
+	    } else if (xsltUTF8Charcmp(*format, self->permille) == 0) {
+		if (info->is_multiplier_set)
+		    return -1;
+		if (info->multiplier != 1000)
+		    return -1;
+		info->is_multiplier_set = TRUE;
+	    }
 	}
-
-	if ((len=xmlUTF8Strsize(*format, 1)) < 1)
+	
+	if ((len=xsltUTF8Size(*format)) < 1)
 	    return -1;
 	count += len;
 	*format += len;
     }
 }
-
+	    
 /**
  * xsltFormatNumberConversion:
  * @self: the decimal format
  * @format: the format requested
  * @number: the value to format
- * @result: the place to output the result
+ * @result: the place to ouput the result
  *
  * format-number() uses the JDK 1.1 DecimalFormat class:
  *
@@ -958,14 +972,15 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
     xmlBufferPtr buffer;
     xmlChar *the_format, *prefix = NULL, *suffix = NULL;
     xmlChar *nprefix, *nsuffix = NULL;
+    xmlChar pchar;
     int	    prefix_length, suffix_length = 0, nprefix_length, nsuffix_length;
     double  scale;
     int	    j, len;
     int     self_grouping_len;
     xsltFormatNumberInfo format_info;
-    /*
+    /* 
      * delayed_multiplier allows a 'trailing' percent or
-     * permille to be treated as suffix
+     * permille to be treated as suffix 
      */
     int		delayed_multiplier = 0;
     /* flag to show no -ve format present for -ve number */
@@ -973,11 +988,6 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
     /* flag to show error found, should use default format */
     char	found_error = 0;
 
-    if (xmlStrlen(format) <= 0) {
-	xsltTransformError(NULL, NULL, NULL,
-                "xsltFormatNumberConversion : "
-		"Invalid format (0-length)\n");
-    }
     *result = NULL;
     switch (xmlXPathIsInf(number)) {
 	case -1:
@@ -985,7 +995,7 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 		*result = xmlStrdup(BAD_CAST "-");
 	    else
 		*result = xmlStrdup(self->minusSign);
-	    /* Intentional fall-through */
+	    /* no-break on purpose */
 	case 1:
 	    if ((self == NULL) || (self->infinity == NULL))
 		*result = xmlStrcat(*result, BAD_CAST "Infinity");
@@ -1021,7 +1031,7 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 
     /*
      * First we process the +ve pattern to get percent / permille,
-     * as well as main format
+     * as well as main format 
      */
     prefix = the_format;
     prefix_length = xsltFormatNumberPreSuffix(self, &the_format, &format_info);
@@ -1030,18 +1040,18 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 	goto OUTPUT_NUMBER;
     }
 
-    /*
-     * Here we process the "number" part of the format.  It gets
+    /* 
+     * Here we process the "number" part of the format.  It gets 
      * a little messy because of the percent/per-mille - if that
-     * appears at the end, it may be part of the suffix instead
-     * of part of the number, so the variable delayed_multiplier
-     * is used to handle it
+     * appears at the end, it may be part of the suffix instead 
+     * of part of the number, so the variable delayed_multiplier 
+     * is used to handle it 
      */
     self_grouping_len = xmlStrlen(self->grouping);
     while ((*the_format != 0) &&
 	   (xsltUTF8Charcmp(the_format, self->decimalPoint) != 0) &&
 	   (xsltUTF8Charcmp(the_format, self->patternSeparator) != 0)) {
-
+	
 	if (delayed_multiplier != 0) {
 	    format_info.multiplier = delayed_multiplier;
 	    format_info.is_multiplier_set = TRUE;
@@ -1079,8 +1089,8 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 	    delayed_multiplier = 1000;
 	} else
 	    break; /* while */
-
-	if ((len=xmlUTF8Strsize(the_format, 1)) < 1) {
+	
+	if ((len=xsltUTF8Size(the_format)) < 1) {
 	    found_error = 1;
 	    goto OUTPUT_NUMBER;
 	}
@@ -1089,18 +1099,13 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
     }
 
     /* We have finished the integer part, now work on fraction */
-    if ( (*the_format != 0) &&
-         (xsltUTF8Charcmp(the_format, self->decimalPoint) == 0) ) {
+    if (xsltUTF8Charcmp(the_format, self->decimalPoint) == 0) {
         format_info.add_decimal = TRUE;
-        if ((len = xmlUTF8Strsize(the_format, 1)) < 1) {
-            found_error = 1;
-            goto OUTPUT_NUMBER;
-        }
-	the_format += len;	/* Skip over the decimal */
+	the_format += xsltUTF8Size(the_format);	/* Skip over the decimal */
     }
-
+    
     while (*the_format != 0) {
-
+	
 	if (xsltUTF8Charcmp(the_format, self->zeroDigit) == 0) {
 	    if (format_info.frac_hash != 0) {
 		found_error = 1;
@@ -1115,7 +1120,7 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 		goto OUTPUT_NUMBER;
 	    }
 	    delayed_multiplier = 100;
-	    if ((len = xmlUTF8Strsize(the_format, 1)) < 1) {
+	    if ((len = xsltUTF8Size(the_format)) < 1) {
 	        found_error = 1;
 		goto OUTPUT_NUMBER;
 	    }
@@ -1127,7 +1132,7 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 		goto OUTPUT_NUMBER;
 	    }
 	    delayed_multiplier = 1000;
-	    if  ((len = xmlUTF8Strsize(the_format, 1)) < 1) {
+	    if  ((len = xsltUTF8Size(the_format)) < 1) {
 	        found_error = 1;
 		goto OUTPUT_NUMBER;
 	    }
@@ -1136,7 +1141,7 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 	} else if (xsltUTF8Charcmp(the_format, self->grouping) != 0) {
 	    break; /* while */
 	}
-	if ((len = xmlUTF8Strsize(the_format, 1)) < 1) {
+	if ((len = xsltUTF8Size(the_format)) < 1) {
 	    found_error = 1;
 	    goto OUTPUT_NUMBER;
 	}
@@ -1148,9 +1153,9 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 	}
     }
 
-    /*
-     * If delayed_multiplier is set after processing the
-     * "number" part, should be in suffix
+    /* 
+     * If delayed_multiplier is set after processing the 
+     * "number" part, should be in suffix 
      */
     if (delayed_multiplier != 0) {
 	the_format -= len;
@@ -1160,7 +1165,7 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
     suffix = the_format;
     suffix_length = xsltFormatNumberPreSuffix(self, &the_format, &format_info);
     if ( (suffix_length < 0) ||
-	 ((*the_format != 0) &&
+	 ((*the_format != 0) && 
 	  (xsltUTF8Charcmp(the_format, self->patternSeparator) != 0)) ) {
 	found_error = 1;
 	goto OUTPUT_NUMBER;
@@ -1182,23 +1187,25 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 	}
 	else {
 	    /* Skip over pattern separator (accounting for UTF8) */
-	    the_format = (xmlChar *)xmlUTF8Strpos(format, j + 1);
-	    /*
-	     * Flag changes interpretation of percent/permille
-	     * in -ve pattern
+	    the_format = xmlUTF8Strpos(format, j + 1);
+	    /* 
+	     * Flag changes interpretation of percent/permille 
+	     * in -ve pattern 
 	     */
 	    format_info.is_negative_pattern = TRUE;
 	    format_info.is_multiplier_set = FALSE;
 
 	    /* First do the -ve prefix */
 	    nprefix = the_format;
-	    nprefix_length = xsltFormatNumberPreSuffix(self,
-					&the_format, &format_info);
+	    nprefix_length = xsltFormatNumberPreSuffix(self, 
+	    				&the_format, &format_info);
 	    if (nprefix_length<0) {
 		found_error = 1;
 		goto OUTPUT_NUMBER;
 	    }
 
+	    /* Next skip over the -ve number info */
+	    the_format += prefix_length;
 	    while (*the_format != 0) {
 		if ( (xsltUTF8Charcmp(the_format, (self)->percent) == 0) ||
 		     (xsltUTF8Charcmp(the_format, (self)->permille)== 0) ) {
@@ -1213,7 +1220,7 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 		    delayed_multiplier = 0;
 		else
 		    break; /* while */
-		if ((len = xmlUTF8Strsize(the_format, 1)) < 1) {
+		if ((len = xsltUTF8Size(the_format)) < 1) {
 		    found_error = 1;
 		    goto OUTPUT_NUMBER;
 		}
@@ -1227,7 +1234,7 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 	    /* Finally do the -ve suffix */
 	    if (*the_format != 0) {
 		nsuffix = the_format;
-		nsuffix_length = xsltFormatNumberPreSuffix(self,
+		nsuffix_length = xsltFormatNumberPreSuffix(self, 
 					&the_format, &format_info);
 		if (nsuffix_length < 0) {
 		    found_error = 1;
@@ -1245,19 +1252,18 @@ xsltFormatNumberConversion(xsltDecimalFormatPtr self,
 	     * if -ve prefix/suffix == +ve ones, discard & use default
 	     */
 	    if ((nprefix_length != prefix_length) ||
-		(nsuffix_length != suffix_length) ||
-		((nprefix_length > 0) &&
+	    	(nsuffix_length != suffix_length) ||
+		((nprefix_length > 0) && 
 		 (xmlStrncmp(nprefix, prefix, prefix_length) !=0 )) ||
-		((nsuffix_length > 0) &&
+		((nsuffix_length > 0) && 
 		 (xmlStrncmp(nsuffix, suffix, suffix_length) !=0 ))) {
-		prefix = nprefix;
+	 	prefix = nprefix;
 		prefix_length = nprefix_length;
 		suffix = nsuffix;
 		suffix_length = nsuffix_length;
-	    } /* else {
+	    } else {
 		default_sign = 1;
 	    }
-	    */
 	}
     }
 
@@ -1265,7 +1271,7 @@ OUTPUT_NUMBER:
     if (found_error != 0) {
 	xsltTransformError(NULL, NULL, NULL,
                 "xsltFormatNumberConversion : "
-		"error in format string '%s', using default\n", format);
+		"error in format string, using default\n");
 	default_sign = (number < 0.0) ? 1 : 0;
 	prefix_length = suffix_length = 0;
 	format_info.integer_hash = 0;
@@ -1279,32 +1285,32 @@ OUTPUT_NUMBER:
 
     /* Ready to output our number.  First see if "default sign" is required */
     if (default_sign != 0)
-	xmlBufferAdd(buffer, self->minusSign, xmlUTF8Strsize(self->minusSign, 1));
+	xmlBufferAdd(buffer, self->minusSign, xsltUTF8Size(self->minusSign));
 
     /* Put the prefix into the buffer */
-    for (j = 0; j < prefix_length; ) {
-	if (*prefix == SYMBOL_QUOTE)
-            prefix++;
-        len = xmlUTF8Strsize(prefix, 1);
-        xmlBufferAdd(buffer, prefix, len);
-        prefix += len;
-        j += len;
+    for (j = 0; j < prefix_length; j++) {
+	if ((pchar = *prefix++) == SYMBOL_QUOTE) {
+	    len = xsltUTF8Size(prefix);
+	    xmlBufferAdd(buffer, prefix, len);
+	    prefix += len + 1;	/* skip the ending quote */
+	    j += len - 1;	/* 'for' will increment by 1 */
+	} else
+	    xmlBufferAdd(buffer, &pchar, 1);
     }
 
     /* Next do the integer part of the number */
     number = fabs(number) * (double)format_info.multiplier;
     scale = pow(10.0, (double)(format_info.frac_digits + format_info.frac_hash));
     number = floor((scale * number + 0.5)) / scale;
-    if ((self->grouping != NULL) &&
+    if ((self->grouping != NULL) && 
         (self->grouping[0] != 0)) {
-        int gchar;
-
+	
 	len = xmlStrlen(self->grouping);
-	gchar = xsltGetUTF8Char(self->grouping, &len);
+	pchar = xsltGetUTF8Char(self->grouping, &len);
 	xsltNumberFormatDecimal(buffer, floor(number), self->zeroDigit[0],
 				format_info.integer_digits,
 				format_info.group,
-				gchar, len);
+				pchar, len);
     } else
 	xsltNumberFormatDecimal(buffer, floor(number), self->zeroDigit[0],
 				format_info.integer_digits,
@@ -1321,20 +1327,20 @@ OUTPUT_NUMBER:
     /* Add leading zero, if required */
     if ((floor(number) == 0) &&
 	(format_info.integer_digits + format_info.frac_digits == 0)) {
-        xmlBufferAdd(buffer, self->zeroDigit, xmlUTF8Strsize(self->zeroDigit, 1));
+        xmlBufferAdd(buffer, self->zeroDigit, xsltUTF8Size(self->zeroDigit));
     }
 
     /* Next the fractional part, if required */
     if (format_info.frac_digits + format_info.frac_hash == 0) {
         if (format_info.add_decimal)
-	    xmlBufferAdd(buffer, self->decimalPoint,
-			 xmlUTF8Strsize(self->decimalPoint, 1));
+	    xmlBufferAdd(buffer, self->decimalPoint, 
+	    		 xsltUTF8Size(self->decimalPoint));
     }
     else {
       number -= floor(number);
 	if ((number != 0) || (format_info.frac_digits != 0)) {
 	    xmlBufferAdd(buffer, self->decimalPoint,
-			 xmlUTF8Strsize(self->decimalPoint, 1));
+	    		 xsltUTF8Size(self->decimalPoint));
 	    number = floor(scale * number + 0.5);
 	    for (j = format_info.frac_hash; j > 0; j--) {
 		if (fmod(number, 10.0) >= 1.0)
@@ -1347,13 +1353,14 @@ OUTPUT_NUMBER:
 	}
     }
     /* Put the suffix into the buffer */
-    for (j = 0; j < suffix_length; ) {
-	if (*suffix == SYMBOL_QUOTE)
-            suffix++;
-        len = xmlUTF8Strsize(suffix, 1);
-        xmlBufferAdd(buffer, suffix, len);
-        suffix += len;
-        j += len;
+    for (j = 0; j < suffix_length; j++) {
+	if ((pchar = *suffix++) == SYMBOL_QUOTE) {
+            len = xsltUTF8Size(suffix);
+	    xmlBufferAdd(buffer, suffix, len);
+	    suffix += len + 1;  /* skip the ending quote */
+	    j += len - 1;       /* 'for' will increment by 1 */
+	} else
+	    xmlBufferAdd(buffer, &pchar, 1);
     }
 
     *result = xmlStrdup(xmlBufferContent(buffer));
